﻿using System;
using SDX.Compiler;
using Mono.Cecil;
using Mono.Cecil.Cil;
using System.Linq;
using System.Collections.Generic;

public class MinWellnessZero : IPatcherMod
{
    public bool Link(ModuleDefinition gameModule, ModuleDefinition modModule)
    {
        InsertGameStageDefinitionRead(gameModule, modModule);
        InsertMinWellnessZeroDeadMode(gameModule, modModule);
        InsertMinWellnessZeroDeadMode2(gameModule, modModule);
        InsertMinWellnessZeroEntityStatsTick(gameModule, modModule);
        return true;
    }
    public enum InsertMinWellnessZeroDeadModeState
    {
        START,
        FINDREMOVESTART,
        FINDREMOVEEND,

        FINDFLOATINITINSRETPOS,
        FINDJUMPTOUPDATEPROG,
        FINDWELLNESS,
        FINDCALLINSRETPOS,

    }
    public enum InsertMinWellnessZeroEntityStatsState
    {
        START,
        FINDVARIABLE,
        FINDINSRETPOS,//FINDISINST,

    }
    //change NetPackagePlayerStats.ProcessPackage if (this.XL > 50) to this.XL > 0
    public void ReplaceNetPackagePlayerStatsProcessPackage50to0(ModuleDefinition gameModule)
    {
        TypeDefinition netPackagePlayerStats = gameModule.Types.First(d => d.Name == "NetPackagePlayerStats");
        MethodDefinition processPackage = netPackagePlayerStats.Methods.First(d => d.Name == "ProcessPackage");

        foreach (Instruction inst in processPackage.Body.Instructions)
        {
            if (inst.OpCode == OpCodes.Ldc_I4_S && 50 == (sbyte)inst.Operand)
            {
                sbyte change = 0;
                inst.Operand = change;
                break;
            }

        }
    }

    public void InsertMinWellnessZeroEntityStatsTick(ModuleDefinition gameModule, ModuleDefinition modModule)
    {
        TypeDefinition entityStats = gameModule.Types.First(d => d.Name == "EntityStats");
        MethodDefinition tick = entityStats.Methods.First(d => d.Name == "Tick");
        MethodDefinition getEntity = entityStats.Methods.First(d => d.Name == "get_Entity");
        TypeDefinition entityPlayer = gameModule.Types.First(d => d.Name == "EntityPlayer");
        FieldDefinition wellnessFld = entityStats.Fields.First(d => d.Name == "Wellness");
        TypeDefinition stat = gameModule.Types.First(d => d.Name == "Stat");
        MethodDefinition statGetValue = stat.Methods.First(d => d.Name == "get_Value");


        VariableDefinition vd = null;
        Instruction insertPos = null;
        Instruction jumpToPos = null;

        InsertMinWellnessZeroEntityStatsState state = InsertMinWellnessZeroEntityStatsState.START;
        foreach (Instruction inst in tick.Body.Instructions)
        {
            if (state == InsertMinWellnessZeroEntityStatsState.FINDINSRETPOS)
            {
                break;

            }
            switch (state)
            {
                case InsertMinWellnessZeroEntityStatsState.START:
                    if (inst.OpCode == OpCodes.Stloc_S && inst.Operand.GetType() == typeof(VariableDefinition) && ((VariableDefinition)inst.Operand).ToString() == "V_8")
                    {
                        vd = (VariableDefinition)inst.Operand;
                        state = InsertMinWellnessZeroEntityStatsState.FINDVARIABLE;

                    }
                    break;
                case InsertMinWellnessZeroEntityStatsState.FINDVARIABLE:
                    if (inst.OpCode == OpCodes.Isinst && inst.Next.OpCode == OpCodes.Brfalse_S)
                    {
                        //inst IL_0239: isinst EntityPlayer opcode isinst operand EntityPlayer
                        //inst IL_023e: brfalse.s IL_0266 opcode brfalse.s operand IL_0266: ldarg.0
                        insertPos = inst.Next.Next;
                        jumpToPos = (Instruction)inst.Next.Operand;
                        state = InsertMinWellnessZeroEntityStatsState.FINDINSRETPOS;

                    }
                    break;
            }
        }
        ILProcessor proc1 = tick.Body.GetILProcessor();
        proc1.InsertBefore(insertPos, Instruction.Create(OpCodes.Ldc_R4, 0f));
        proc1.InsertBefore(insertPos, Instruction.Create(OpCodes.Stloc_S, vd));
        proc1.InsertBefore(insertPos, Instruction.Create(OpCodes.Ldc_R4, 0f));
        proc1.InsertBefore(insertPos, Instruction.Create(OpCodes.Ldarg_0));
        proc1.InsertBefore(insertPos, Instruction.Create(OpCodes.Ldfld, wellnessFld));
        proc1.InsertBefore(insertPos, Instruction.Create(OpCodes.Callvirt, statGetValue));
        proc1.InsertBefore(insertPos, Instruction.Create(OpCodes.Beq_S, jumpToPos));

    }

    public void InsertMinWellnessZeroDeadMode(ModuleDefinition gameModule, ModuleDefinition modModule)
    {
        TypeDefinition entityPlayer = gameModule.Types.First(d => d.Name == "EntityPlayer");
        MethodDefinition setAlive = entityPlayer.Methods.First(d => d.Name == "SetAlive");
        Instruction removeStart = null;
        int removeStartPos = 0;
        Instruction removeEnd = null;
        int removeEndPos = 0;
        Instruction insertPos = null;
        InsertMinWellnessZeroDeadModeState state = InsertMinWellnessZeroDeadModeState.START;
        for (int i = 0;i <  setAlive.Body.Instructions.Count;i++)
        {
            Instruction inst = setAlive.Body.Instructions[i];
            switch (state)
            {
                case InsertMinWellnessZeroDeadModeState.START:
                    if (inst.OpCode == OpCodes.Callvirt && inst.Operand.GetType() == typeof(MethodDefinition) && ((MethodDefinition)inst.Operand).FullName == "System.Void Skills::ModifyValue(Skill/Effects,System.Single&,Skill)")
                    {
                        removeStart = inst.Next;
                        removeStartPos = i + 1;
                        state = InsertMinWellnessZeroDeadModeState.FINDREMOVESTART;
                    }
                    break;
                case InsertMinWellnessZeroDeadModeState.FINDREMOVESTART:
                    if (inst.OpCode == OpCodes.Callvirt && inst.Operand.GetType() == typeof(MethodDefinition) && ((MethodDefinition)inst.Operand).FullName == "System.Void Stat::set_Value(System.Single)")
                    {
                        removeEnd = inst;
                        removeEndPos = i;
                        insertPos = inst.Next;
                        state = InsertMinWellnessZeroDeadModeState.FINDREMOVESTART;
                    }
                    break;
            }

        }

        ILProcessor proc1 = setAlive.Body.GetILProcessor();
        for (int i = removeEndPos; i >= removeStartPos ;i--)
        {
            Instruction inst = setAlive.Body.Instructions[i];
            proc1.Remove(inst);
        }


        proc1.InsertBefore(insertPos, Instruction.Create(OpCodes.Ldarg_0));

        TypeDefinition gameStageDef = gameModule.Types.First(d => d.Name == "GameStageDefinition");
        FieldDefinition gameOverModeField = gameStageDef.Fields.First(d => d.Name == "GameOverMode");


        proc1.InsertBefore(insertPos, Instruction.Create(OpCodes.Ldsfld, gameOverModeField));


        proc1.InsertBefore(insertPos, Instruction.Create(OpCodes.Ldloc_1));

        TypeDefinition minWellnessZeroDeadMode = modModule.Types.First(d => d.Name == "MinWellnessZeroDeadMode");
        MethodDefinition execDeadMode = minWellnessZeroDeadMode.Methods.First(d => d.Name == "ExecDeadMode");
        MethodReference execDeadModeRef = gameModule.Import(execDeadMode);

        proc1.InsertBefore(insertPos, Instruction.Create(OpCodes.Call, execDeadModeRef));

    }
    public void InsertMinWellnessZeroDeadMode2(ModuleDefinition gameModule, ModuleDefinition modModule)
    {
        TypeDefinition entityAlive = gameModule.Types.First(d => d.Name == "EntityAlive");
        MethodDefinition setAlive = entityAlive.Methods.First(d => d.Name == "SetAlive");

        Instruction floatInitInsertPos = null;
        Instruction CallInsertPos = null;
        InsertMinWellnessZeroDeadModeState state = InsertMinWellnessZeroDeadModeState.START;
        FieldDefinition wellnessFld = null;
        MethodDefinition statGetValue = null;
        MethodDefinition getStats = null;
        Instruction JumpToUpdateProg = null;
        Instruction InsertWellnessCheckPos = null;
        foreach (Instruction inst in setAlive.Body.Instructions)
        {
            switch (state)
            {
                case InsertMinWellnessZeroDeadModeState.START:
                    if (inst.OpCode == OpCodes.Callvirt && inst.Operand.GetType() == typeof(MethodDefinition) && ((MethodDefinition)inst.Operand).FullName == "System.Void EntityStats::ResetStats()")
                    {
                        floatInitInsertPos = inst;
                        state = InsertMinWellnessZeroDeadModeState.FINDFLOATINITINSRETPOS;
                    }
                    break;
                case InsertMinWellnessZeroDeadModeState.FINDFLOATINITINSRETPOS:
                    if (inst.OpCode == OpCodes.Brfalse_S)
                    {
                        JumpToUpdateProg = inst;
                        InsertWellnessCheckPos = (Instruction)inst.Operand;
                        state = InsertMinWellnessZeroDeadModeState.FINDJUMPTOUPDATEPROG;
                    }
                    break;
                case InsertMinWellnessZeroDeadModeState.FINDJUMPTOUPDATEPROG:
                    if (inst.OpCode == OpCodes.Call && inst.Operand.GetType() == typeof(MethodDefinition) && ((MethodDefinition)inst.Operand).FullName == "EntityStats EntityAlive::get_Stats()")
                    {
                        getStats = (MethodDefinition)inst.Operand;
                    }
                    else if (inst.OpCode == OpCodes.Ldfld && inst.Operand.GetType() == typeof(FieldDefinition) && ((FieldDefinition)inst.Operand).FullName == "Stat EntityStats::Wellness")
                    {
                        wellnessFld = (FieldDefinition)inst.Operand;
                        statGetValue = (MethodDefinition)inst.Next.Operand;
                        state = InsertMinWellnessZeroDeadModeState.FINDWELLNESS;
                    }
                    break;
                case InsertMinWellnessZeroDeadModeState.FINDWELLNESS:
                    if (inst.OpCode == OpCodes.Callvirt && inst.Operand.GetType() == typeof(MethodDefinition) && ((MethodDefinition)inst.Operand).FullName == "System.Void EntityStats::UpdatePlayerProgression()")
                    {
                        CallInsertPos = inst.Next;
                        state = InsertMinWellnessZeroDeadModeState.FINDCALLINSRETPOS;
                    }
                    break;
            }
        }

        ILProcessor proc1 = setAlive.Body.GetILProcessor();
        //float value init 1000f
        proc1.InsertBefore(floatInitInsertPos, Instruction.Create(OpCodes.Ldc_R4, 1000f));
        proc1.InsertBefore(floatInitInsertPos, Instruction.Create(OpCodes.Stloc_1));


        TypeDefinition gameStageDef = gameModule.Types.First(d => d.Name == "GameStageDefinition");
        FieldDefinition gameOverModeField = gameStageDef.Fields.First(d => d.Name == "GameOverMode");


        Instruction newJumpPos = Instruction.Create(OpCodes.Ldarg_0);
        proc1.InsertBefore(CallInsertPos, newJumpPos);
        proc1.InsertBefore(CallInsertPos, Instruction.Create(OpCodes.Ldsfld, gameOverModeField));
        proc1.InsertBefore(CallInsertPos, Instruction.Create(OpCodes.Ldloc_1));

        TypeDefinition minWellnessZeroDeadMode = modModule.Types.First(d => d.Name == "MinWellnessZeroDeadMode");
        MethodDefinition execDeadMode = minWellnessZeroDeadMode.Methods.First(d => d.Name == "ExecDeadMode2");
        MethodReference execDeadModeRef = gameModule.Import(execDeadMode);

        proc1.InsertBefore(CallInsertPos, Instruction.Create(OpCodes.Call, execDeadModeRef));


    }



    public void InsertGameStageDefinitionRead(ModuleDefinition gameModule, ModuleDefinition modModule)
    {
        TypeDefinition classDef = gameModule.Types.First(d => d.Name == "GameStagesFromXml");
        MethodDefinition methodDef = classDef.Methods.First(d => d.Name == "Load");

        TypeDefinition gameStageDef = gameModule.Types.First(d => d.Name == "GameStageDefinition");        
        FieldDefinition gameOverModeField = gameStageDef.Fields.First(d => d.Name == "GameOverMode");

        Instruction insertPosition = null;
        foreach (Instruction inst in methodDef.Body.Instructions)
        {

            if (inst.OpCode == OpCodes.Stsfld && inst.Operand.GetType() == typeof(FieldDefinition) && ((FieldDefinition)inst.Operand).FullName == "System.Single GameStageDefinition::DaysAliveDifficultyBonus")
            {
                insertPosition = inst.Next;

            }
        }
        List<Instruction> jumpList = new List<Instruction>();
        foreach (Instruction inst in methodDef.Body.Instructions)
        {
            if (inst.OpCode == OpCodes.Brfalse_S && inst.Operand.GetType() == typeof(Instruction) && inst.Operand == insertPosition)
            {
                jumpList.Add(inst);
            }
        }

        ILProcessor proc1 = methodDef.Body.GetILProcessor();
        Instruction newJumpTo = Instruction.Create(OpCodes.Ldloc_S, proc1.Body.Variables[4]);
        proc1.InsertBefore(insertPosition, newJumpTo);//V_4 is xmlelement
        proc1.InsertBefore(insertPosition, Instruction.Create(OpCodes.Ldsflda, gameOverModeField));//ref GameOverMode

        TypeDefinition minWellnessZeroDeadMode = modModule.Types.First(d => d.Name == "MinWellnessZeroDeadMode");
        MethodDefinition setAddAttribute = minWellnessZeroDeadMode.Methods.First(d => d.Name == "SetAddAttribute");

        if (setAddAttribute == null)
        {
            Console.WriteLine("InsertCallBuffsFromXmlAdd setAddAttribute == null");
        }
        MethodReference setAddAttributeRef = gameModule.Import(setAddAttribute);
        Console.WriteLine("InsertCallBuffsFromXmlAdd setAddAttribute find");

        proc1.InsertBefore(insertPosition, Instruction.Create(OpCodes.Call, setAddAttributeRef));

        foreach (Instruction jump in jumpList)
        {
            jump.Operand = newJumpTo;
        }
    }

    public bool Patch(ModuleDefinition module)
    {
        Console.WriteLine("==MinWellnessZero Patcher===");
        Update70F(module);
        AddGameStageDefinitionField(module);
        return true;
    }
    public void AddGameStageDefinitionField(ModuleDefinition module)
    {
        TypeDefinition multiBuffClass = module.Types.First(d => d.Name == "GameStageDefinition");
        TypeReference tr = module.Import(typeof(byte));
        AddFieldDefinition(multiBuffClass, "GameOverMode", FieldAttributes.Public| FieldAttributes.Static, tr);

    }

    public void AddFieldDefinition(TypeDefinition classDef, string fieldName, FieldAttributes fieldAttribute, TypeReference fieldType)
    {
        FieldDefinition fd = new FieldDefinition(fieldName, fieldAttribute, fieldType);
        classDef.Fields.Add(fd);

    }
    public void AddFieldDefinition(TypeDefinition classDef, string fieldName, FieldAttributes fieldAttribute, TypeReference fieldType, object constantvalue)
    {
        FieldDefinition fd = new FieldDefinition(fieldName, fieldAttribute, fieldType);
        fd.Constant = constantvalue;
        //        fd.
        classDef.Fields.Add(fd);

    }

    private void Update70F(ModuleDefinition module)
    {

        Update70F(module, "EntityAlive", "SetAlive", 1);
        Update70F(module, "EntityPlayer", "SetAlive", 1);
        Update70FConstantField(module, "EntityStats", "MinWellness");
        ReplaceNetPackagePlayerStatsProcessPackage50to0(module);
    }

    public void Update70F(ModuleDefinition module, String strModuleName, String strMethodName, int maxCounter)
    {
        var varModule = module.Types.First(d => d.Name == strModuleName);
        var varMethod = varModule.Methods.First(d => d.Name == strMethodName);
        var instructions = varMethod.Body.Instructions;

        int count = 0;

        foreach (var i in instructions)
        {
            if ((i.OpCode == OpCodes.Ldc_R4) && ((float)i.Operand == 70f))
            {
                i.Operand = 0f;

                if (++count == maxCounter)
                    break;
            }
        }
    }
    public void Update70FConstantField(ModuleDefinition module, String strModuleName, String strFieldName)
    {
        var varModule = module.Types.First(d => d.Name == strModuleName);
        FieldDefinition field = varModule.Fields.First(d => d.Name == strFieldName);
        field.Constant = 0f;

    }
}
